// Copyright 2020 Tencent
// SPDX-License-Identifier: BSD-3-Clause

#ifndef NCNN_LAYER_SHADER_TYPE_H
#define NCNN_LAYER_SHADER_TYPE_H

namespace ncnn {

namespace LayerShaderType {
enum LayerShaderType
{
#include "layer_shader_type_enum.h"
};
} // namespace LayerShaderType

} // namespace ncnn

#endif // NCNN_LAYER_SHADER_TYPE_H
